#!/usr/bin/env python

import xml.etree.ElementTree as ET
import sqlite3
import logging
import textwrap
from distutils.file_util import copy_file
import argparse

# Namespace of GPX. See http://www.topografix.com/gpx.asp for specs.
# Note: GPX 1.1 has namespace "http://www.topografix.com/GPX/1/1"
gpxNS = 'http://www.topografix.com/GPX/1/0'
gpxPrefix = 'gpx'

nokiaMapsDbFile = "/home/user/.cities/NokiaMaps.db"

ET._namespace_map[gpxNS] = gpxPrefix

class POI:
    def __init__(self, lat=0.0, lon=0.0, name=None, desc=None, object_id=None):
        self.name = name
        self.desc = desc
        self.lat = float(lat)
        self.lon = float(lon)
        self.object_id = object_id

    def __str__(self):
        return "lat: %f; lon: %f; name: '%s'; desc: '%s'; object_id: %s" % (self.lat, self.lon, self.name, self.desc, self.object_id)


def load_gpx(gpx_file):
    '''
    Opens GPX file, reads all waypoints and returns them as a list of
    POI objects.
    '''
    root = ET.parse(gpx_file)

    waypoints = root.findall(".//{%s}wpt" % gpxNS)

    pois = []

    for waypoint in waypoints:
        args = {'lat': waypoint.attrib['lat'], 'lon': waypoint.attrib['lon']}
        node =  waypoint.find('{%s}name' % gpxNS)
        if node != None:
            args['name'] = node.text
        node =  waypoint.find('{%s}desc' % gpxNS)
        if node != None:
            args['desc'] = node.text

        pois.append(POI(**args))
                           
    return pois


def save_gpx(pois, gpx_file):
    # define root note, with default namespace set in a hackerish way. But hey, it does the job.
    root = ET.Element("gpx", {'creator': "Python Nokia N9 FavMan", 'xmlns': gpxNS})    

    for poi in pois:
        wpt = ET.SubElement(root, "wpt", {'lat': str(poi.lat), 'lon': str(poi.lon)})
        if poi.name != None:
            ET.SubElement(wpt, "name").text = poi.name
        if poi.desc != None:
            ET.SubElement(wpt, "desc").text = poi.desc
    
            
    ET.ElementTree(root).write(gpx_file, encoding="UTF-8")

def load_db(db_file):
    '''
    Opens the database and obtain a list of all objects having a
    position, and get their object_id, their name and description attribute. Returns a
    list of POI objects. 
    '''
    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()

    cursor.execute(
        """
        select p.latitude, p.longitude, t1.value_text as name, t2.value_text as description, o.local_id as object_id
        from object o
        join position p on p.parent_object = o.local_id
        left join text t1 on t1.parent_object = o.local_id and t1.parent_attribute in (
            select a2.id from attribute a1 join attribute a2 on a1.id = a2.parent_attribute and a1.class_name = 'metadata' and a2.name = 'name'
        )
        left join text t2 on t2.parent_object = o.local_id and t2.parent_attribute in (
            select a2.id from attribute a1 join attribute a2 on a1.id = a2.parent_attribute and a1.class_name = 'metadata' and a2.name = 'description'
        )
        where o.class_name='location'
        """)    

    pois = [POI(lat=row[0], lon=row[1], name=row[2], desc=row[3], object_id=row[4]) for row in cursor.fetchall()]
    conn.close()
    
    return pois


def save_db(pois, db_file):
    '''
    Stores the list of POIs in the database
    '''
    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()

    counter = 0

    for poi in pois:        
        if poi.object_id == None:
            # insert object
            cursor.execute("insert into object (class_name, modified) values ('location', 1)")
            poi.object_id = cursor.lastrowid

            # add position
            cursor.execute("insert into attribute (class_name, name, parent_object, parent_attribute) values ('position', 'position', ?, 0)", (poi.object_id,))
            cursor.execute("insert into position(latitude, longitude, parent_object, parent_attribute) values (?, ?, ?, ?)", (poi.lat, poi.lon, poi.object_id, cursor.lastrowid))

            # add metadata attribute
            cursor.execute("insert into attribute (class_name, name, parent_object, parent_attribute) values ('metadata', 'meta', ?, 0)", (poi.object_id,))
            metaID = cursor.lastrowid

            # add name
            if poi.name != None:
                cursor.execute("insert into attribute (class_name, name, parent_object, parent_attribute) values ('text', 'name', ?, ?)", (poi.object_id, metaID))
                nameID = cursor.lastrowid
                cursor.execute("insert into text (name, type, value_text, parent_object, parent_attribute) values ('text', 8, ?, ?, ?)", (poi.name, poi.object_id, nameID))

            # add description
            if poi.name != None:
                cursor.execute("insert into attribute (class_name, name, parent_object, parent_attribute) values ('text', 'description', ?, ?)", (poi.object_id, metaID))
                descID = cursor.lastrowid
                cursor.execute("insert into text (name, type, value_text, parent_object, parent_attribute) values ('text', 8, ?, ?, ?)", (poi.desc, poi.object_id, descID))

            conn.commit()
            counter+=1
        else:
            logging.warning("Updating favorites not yet supported; only new favorites can be added.");

    conn.close()
    
    return counter
    

def delete_db(pois, db_file):
    '''
    Deletes given POIs from the database. However, only POIs with their
    local_id set are deleted. Therefor, this function is best used in
    conjunction with the load_db function.    
    '''
    conn = sqlite3.connect(db_file)

    for poi in pois:
        if poi.local_id != None:
            # Note: delete will trigger the database triggers, which will clean up the referred tables
            conn.execute("delete from objects where local_id = ?", (poi.local_id,))

    conn.commit()
    conn.close()
    
             
# Command line interface handling

parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent("""\
        FavMan v1.0.0. Manages favorites/POIs in the Nokia N9 Maps database.
        Public domain by Randy Simons.

        WARNING

        This is experimental software. Be sure to backup your Nokia Maps database, located in /home/user/.cities/NokiaMaps.db, before using this tool.
        """),
    epilog = textwrap.dedent("""\
        Notes:
         * It is yet unknown what will happen if you sync imported favorites. It probably won't blow up your phone, but please do make a backup!
        
        Tips:        
         * Only GPX v1.0 files supported. Favorities must/will be stored as waypoints.
         * Use GPSBabel (gpsbabel.org) to convert various type of geolocation data, including GPX and Nokia's LMX.
         * Nokia Maps database is stored in /home/user/.cities/NokiaMaps.db         
         * Nokia Maps will recreate a fresh datbase if the original database cannot be found. (e.g. renamed, deleted)
         * Use Sqlite3Explorer (http://www.singular.gr/sqlite/, download dll from http://sqlite.org/download.html) to inspect Nokia Maps database.
        
        Examples:
        %(prog)s -loaddb
        Opens default database and shows all favorites.

        %(prog)s -loadgpx
        Loads waypoints from favorites.gpx and displays them on screen. Use this to verify if the gpx file can be read correctly.

        %(prog)s -loaddb -savegpx
        Opens default database and exports favorites to POIs.gpx

        %(prog)s -loaddb NokiaMapsCopy.db -savegpx export.gpx
        Opens database NokiaMapsCopy.db and exports favorites to export.gpx

        %(prog)s -loadgpx -savedb 
        Loads waypoints from favorites.gpx and stores them as favorites in the database.
        
        %(prog)s -loadgpx import.gpx -savedb export.db
        Loads waypoints from import.gpx and stores them as favorites in the database export.db. The original database will remain unchanged.
        """)
    )

loaders = parser.add_mutually_exclusive_group(required=True)
loaders.add_argument("-loaddb", nargs='?', metavar="file", help="Load favorites from database. Default if no file specified: %(const)s", const=nokiaMapsDbFile)
loaders.add_argument("-loadgpx", nargs='?', metavar="file", help="Load favorites from GPX file. Default if no file specified: %(const)s", const="favorites.gpx")

savers = parser.add_mutually_exclusive_group()
savers.add_argument("-savedb", nargs='?', metavar="file", help="Save favorites to database. Default if no file specified: %(const)s", const=nokiaMapsDbFile)
savers.add_argument("-savegpx", nargs='?', metavar="file", help="Save favorites to GPX file Default if no file specified: %(const)s", const="favorites.gpx")

args = parser.parse_args()

# Loading
if args.loaddb != None:
    pois = load_db(args.loaddb)
else:
    pois = load_gpx(args.loadgpx)

print("Loaded %d favorites." % len(pois))

# Saving
if args.savedb != None:
    if nokiaMapsDbFile != args.savedb:
        copy_file(nokiaMapsDbFile, args.savedb)
        print("Copied original database to %s." % args.savedb)
    number = save_db(pois, args.savedb)
    print("Added %d favorites in database %s." % (number, args.savedb))
elif args.savegpx != None:
    save_gpx(pois, args.savegpx)
    print("Saved favorites in %s." % args.savegpx)
else:
    print("No save option specified. Displaying loaded favorites instead:")
    for poi in pois:
        print(poi)
